@@ -1,6 +1,6 @@
-;;; geiser-doc.el -- accessing scheme-provided documentation
+;;; geiser-doc.el --- Accessing scheme-provided documentation  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016 Jose Antonio Ortega Ruiz
+;; Copyright (C) 2009-2016, 2021-2022, 2024 Jose Antonio Ortega Ruiz
 
 ;; This program is free software; you can redistribute it and/or
 ;; modify it under the terms of the Modified BSD License. You should
@@ -10,6 +10,7 @@
 ;; Start date: Sat Feb 14, 2009 14:09
 
 
+;;; Code:
 
 (require 'geiser-edit)
 (require 'geiser-impl)
@@ -23,6 +24,7 @@
 (require 'geiser-base)
 
 (require 'button)
+(eval-when-compile (require 'subr-x))
 
 
 ;;; Customization:
@@ -49,6 +51,18 @@ help buffer, after collecting the associated signature and
 docstring. You can provide an alternative function for displaying
 help (e.g. browse an HTML page) implementing this method.")
 
+(geiser-impl--define-caller geiser-doc--display-docstring
+    display-docstring (ret)
+  "This method receives the result of calling the geiser scheme
+procedure symbol-documentation and should display it in the
+current buffer.  By default, geiser looks for the value of the
+key docstring in the result, assumed to be an alist, and inserts
+it verbatim at point if it's a string.  Providing an
+implementation of this method may be useful if displaying the
+info returned by the scheme side (display-docstring) needs more
+elaboration on emacs' side.  This method should return a truthy
+value if the default action should be skipped.")
+
 
 ;;; Documentation browser history:
 
@@ -62,10 +76,6 @@ help (e.g. browse an HTML page) implementing this method.")
 
 (setq geiser-doc--history (geiser-doc--make-history))
 
-(defvar session-globals-exclude)
-(eval-after-load "session"
-  '(add-to-list 'session-globals-exclude 'geiser-doc--history))
-
 (defsubst geiser-doc--history-current ()
   (car geiser-doc--history))
 
@@ -127,11 +137,10 @@ help (e.g. browse an HTML page) implementing this method.")
       (with--geiser-implementation impl
         (if (null target)
             (geiser-doc-module module impl)
-          (let ((geiser-eval--get-module-function (lambda (x) module)))
+          (let ((geiser-eval--get-module-function (lambda (_) module)))
             (geiser-doc-symbol target module impl)))))))
 
-(make-variable-buffer-local
- (defvar geiser-doc--buffer-link nil))
+(defvar-local geiser-doc--buffer-link nil)
 
 (defsubst geiser-doc--implementation ()
   (geiser-doc--link-impl geiser-doc--buffer-link))
@@ -140,66 +149,68 @@ help (e.g. browse an HTML page) implementing this method.")
   (let ((link (button-get button 'geiser-link)))
     (when link (geiser-doc--follow-link link))))
 
-(define-button-type 'geiser-doc--button
+(define-button-type 'geiser-doc
   'action 'geiser-doc--button-action
   'follow-link t)
 
 (defun geiser-doc--make-module-button (beg end module impl)
   (let ((link (geiser-doc--make-link nil module impl))
         (help (format "Help for module %s" module)))
-    (make-text-button beg end :type 'geiser-doc--button
+    (make-text-button beg end :type 'geiser-doc
                       'face 'geiser-font-lock-doc-link
                       'geiser-link link
                       'help-echo help)))
 
 (defun geiser-doc--insert-button (target module impl &optional sign)
-  (let ((link (geiser-doc--make-link target module impl))
-        (text (format "%s" (or (and sign
-                                    (geiser-autodoc--str* sign))
-                               target
-                               module)))
-        (help (format "%smodule %s"
-                      (if target (format "%s in " target) "")
-                      (or module "<unknown>"))))
+  (let* ((link (geiser-doc--make-link target module impl))
+         (sign (when sign (if (listp sign) sign (list target))))
+         (text (format "%s" (or (and sign (geiser-autodoc--str* sign))
+                                target
+                                module)))
+         (help (format "%smodule %s"
+                       (if target (format "%s in " target) "")
+                       (or module "<unknown>"))))
     (insert-text-button text
-                        :type 'geiser-doc--button
+                        :type 'geiser-doc
                         'face 'geiser-font-lock-doc-link
                         'geiser-link link
                         'help-echo help)))
 
-(defun geiser-doc--xbutton-action (button)
-  (when geiser-doc--buffer-link
-    (let ((kind (or (button-get button 'x-kind) 'source))
-          (target (geiser-doc--link-target geiser-doc--buffer-link))
-          (module (geiser-doc--link-module geiser-doc--buffer-link))
-          (impl (geiser-doc--link-impl geiser-doc--buffer-link)))
-      (with--geiser-implementation impl
-        (cond ((eq kind 'source)
-               (if target (geiser-edit-symbol target nil (point-marker))
-                 (geiser-edit-module module)))
-              ((eq kind 'manual)
-               (geiser-doc--external-help impl
-                                          (or target module)
-                                          module)))))))
-
-(define-button-type 'geiser-doc--xbutton
-  'action 'geiser-doc--xbutton-action
+(define-button-type 'geiser-doc-source
+  'action 'geiser-doc-goto-source
   'face 'geiser-font-lock-doc-button
+  'help-echo "Go to definition"
   'follow-link t)
 
-(defun geiser-doc--insert-xbutton (&optional manual)
-  (let ((label (if manual "[manual]" "[source]"))
-        (help (if manual "Look up in Scheme manual" "Go to definition")))
-  (insert-text-button label
-                      :type 'geiser-doc--xbutton
-                      'help-echo help
-                      'x-kind (if manual 'manual 'source))))
+(defun geiser-doc-goto-source (&rest _)
+  "Go to the definition of this item."
+  (interactive)
+  (when-let (link geiser-doc--buffer-link)
+    (with--geiser-implementation (geiser-doc--link-impl link)
+      (if-let (target (geiser-doc--link-target link))
+          (geiser-edit-symbol target nil (point-marker))
+        (geiser-edit-module (geiser-doc--link-module link))))))
 
-(defun geiser-doc--insert-xbuttons (impl)
+(define-button-type 'geiser-doc-manual
+  'action 'geiser-doc-goto-manual
+  'face 'geiser-font-lock-doc-button
+  'help-echo "Look up in Scheme manual"
+  'follow-link t)
+
+(defun geiser-doc-goto-manual (&rest _)
+  "Go to the manual for this item."
+  (interactive)
+  (when-let (link geiser-doc--buffer-link)
+    (let ((tm (geiser-doc--link-target link))
+          (mod (geiser-doc--link-module link))
+          (impl (geiser-doc--link-impl link)))
+      (geiser-doc--external-help impl (or tm mod) mod))))
+
+(defun geiser-doc--insert-doc-buttons (impl)
   (when (geiser-impl--method 'external-help impl)
-    (geiser-doc--insert-xbutton t)
+    (insert-text-button "[manual]" :type 'geiser-doc-manual)
     (insert " "))
-  (geiser-doc--insert-xbutton))
+  (insert-text-button "[source]" :type 'geiser-doc-source))
 
 (defun geiser-doc--insert-nav-button (next)
   (let* ((lnk (if next (geiser-doc--history-next-link)
@@ -252,7 +263,7 @@ help (e.g. browse an HTML page) implementing this method.")
 
 (defun geiser-doc--insert-footer (impl)
   (newline 2)
-  (geiser-doc--insert-xbuttons impl)
+  (geiser-doc--insert-doc-buttons impl)
   (let* ((prev (and (geiser-doc--history-previous-p) 8))
          (nxt (and (geiser-doc--history-next-p) 10))
          (len (max 1 (- (window-width)
@@ -287,9 +298,11 @@ help (e.g. browse an HTML page) implementing this method.")
     map)
   "Keymap for `geiser-doc-mode'.")
 
+(declare-function geiser-repl--switch-to-repl "geiser-repl")
+
 (defun geiser-doc-switch-to-repl ()
   (interactive)
-  (switch-to-geiser nil nil (current-buffer)))
+  (geiser-repl--switch-to-repl))
 
 (geiser-menu--defmenu doc geiser-doc-mode-map
   ("Next link" ("n") forward-button)
@@ -307,6 +320,8 @@ help (e.g. browse an HTML page) implementing this method.")
   --
   ("Edit symbol" ("." "\M-.") geiser-doc-edit-symbol-at-point
    :enable (geiser--symbol-at-point))
+  ("View source" ("s") geiser-doc-goto-source)
+  ("View manual" ("m" "h") geiser-doc-goto-manual)
   --
   ("Kill item" "k" geiser-doc-kill-page "Kill this page")
   ("Clear history" "c" geiser-doc-clean-history)
@@ -324,7 +339,7 @@ help (e.g. browse an HTML page) implementing this method.")
   (setq geiser-eval--get-module-function 'geiser-doc--module)
   (setq buffer-read-only t))
 
-(geiser-popup--define doc "*Geiser documentation*" geiser-doc-mode)
+(geiser-popup--define doc "*Geiser Documentation*" geiser-doc-mode)
 
 
 ;;; Commands:
@@ -347,13 +362,13 @@ help (e.g. browse an HTML page) implementing this method.")
                                                           impl)
                                       impl))))
 
-(defun geiser-doc--render-docstring
-  (docstring symbol &optional module impl)
+(defun geiser-doc--render-docstring (docstring symbol &optional module impl)
   (erase-buffer)
   (geiser-doc--insert-title
    (geiser-autodoc--str* (cdr (assoc "signature" docstring))))
   (newline)
-  (insert (or (cdr (assoc "docstring" docstring)) ""))
+  (or (geiser-doc--display-docstring impl docstring)
+      (insert (or (cdr (assoc "docstring" docstring)) "")))
   (geiser-doc--buttonize-modules impl)
   (setq geiser-doc--buffer-link
         (geiser-doc--history-push (geiser-doc--make-link symbol
@@ -382,6 +397,11 @@ With prefix argument, ask for symbol (with completion)."
                      "Symbol: " (geiser--symbol-at-point)))))
     (when symbol (geiser-doc-symbol symbol))))
 
+(defun geiser-doc-manual-for-symbol (symbol)
+  (geiser-doc--external-help geiser-impl--implementation
+                             symbol
+                             (geiser-eval--get-module)))
+
 (defun geiser-doc-look-up-manual (&optional arg)
   "Look up manual for symbol at point.
 With prefix argument, ask for the lookup symbol (with completion)."
@@ -390,9 +410,7 @@ With prefix argument, ask for the lookup symbol (with completion)."
     (error "No manual available"))
   (let ((symbol (or (and (not arg) (geiser--symbol-at-point))
                     (geiser-completion--read-symbol "Symbol: "))))
-    (geiser-doc--external-help geiser-impl--implementation
-                               symbol
-                               (geiser-eval--get-module))))
+    (geiser-doc-manual-for-symbol symbol)))
 
 (defconst geiser-doc--sections '(("Procedures:" "procs")
                                  ("Syntax:" "syntax")
